/*
 * Copyright 2009-2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var serializer;

function getSerializer() {
    if( serializer == null ) serializer = new XMLSerializer();
    return serializer;
}

//----------------------------------------------------------------

function serializeDocument() {
    removeAllLinks();
    deactivateLocations();
    var scripts = removeScripts();
    var res = getSerializer().serializeToString( document );
    injectScripts( scripts );
    addLocationLinks();
    return res;
}

function serializeDocumentRaw() {
    return getSerializer().serializeToString( document );
}

function injectScripts( scripts ) {
    $( "head" ).append( scripts );
}

function removeScripts() {
    var res = $("#jquery_js").add("#edit_js").add("#edit_css");
    res.remove();
    return res;
}

/**
 * Mark intersections of range and text node range.
 * 
 * new:
 * 
 * <span class="requality_text id_${uuid}">
 *     <a name="${uuid}" id="id_${uuid}" class="requality_id"/>
 *     <span class="requality_link requality_link_(in)?visible">${id}</span>
 *     ...
 * </span>
 * <span class="requality_text id_${uuid}">...</span>
 * 
 * old:
 * 
 * <span class="location_text id_${uuid}"><a name="${uuid}" id="id_${uuid}" class="location_id id_${uuid}"/>...</span>
 * <span class="location_link id_${uuid} location_link_(in)?visible>${id}</span>
 * <span class="location_text id_${uuid}">...</span>
 * <span class="location_text id_${uuid}">...<span class="location_end id_${uuid}"/></span>
 * 
 * @param range    
 * @param textNode
 * @return 1 if intersection of range and textNode is not empty and is not equal to textNode
 * To mark such intersection we add one child node. 
 */
function markTextRange( id, range, textNode ) {
    var r = document.createRange();
    r.selectNode(textNode);
    var res = 0;
    var isStart = false;
    var isEnd = false;
    // r is before range: r.end <= range.start
    if( range.compareBoundaryPoints(Range.END_TO_START,r) >= 0 ) return 0;
    // r is after range: range.end <= r.start
    if( range.compareBoundaryPoints(Range.START_TO_END,r) <= 0 ) return 0;
    switch( range.compareBoundaryPoints(Range.START_TO_START,r) ) {
    case 1:
        r.setStart( range.startContainer, range.startOffset );
        res = 1;
    case 0:
    	isStart = true;
    }
    switch( range.compareBoundaryPoints(Range.END_TO_END,r) ) {
    case -1:
        r.setEnd( range.endContainer, range.endOffset );
        res = 1;
    case 0:
        isEnd = true;
    }
    if( r.toString().trim().length == 0 ) return 0;
    var reqwrap = $("<span/>").attr("class","requality_text id_"+id);
    r.surroundContents( reqwrap.get(0) );
    if( isStart )
    {
        reqwrap.prepend( $("<a/>").attr("name",id).attr("id","id_"+id).attr("class","requality_id") );
    }
//    if( isEnd )
//    {
//        reqwrap.append($("<span/>").attr("class","location_end id_"+id));
//    }
    return res;
}

/**
 * Mark part of 'node'-subtree that is contained in 'range'
 * 
 * @param id - location unique id
 * @param range
 * @param node
 */
function markRange( id, range, node ) {
    var r = document.createRange();

    var childNodesCount = node.childNodes.length;
    if( childNodesCount > 0 )
    {
        for( var i = 0; i < childNodesCount; i++ )
        {
            var child = node.childNodes.item(i);
            r.selectNode(child);
            // r is after range: range.end <= r.start
            if( range.compareBoundaryPoints(Range.START_TO_END,r) <= 0 )
            {
                break;
            }
            
            if( child.nodeType == Node.TEXT_NODE )
            {
                var newChildCount = markTextRange( id, range, child );
                i += newChildCount;
                childNodesCount += newChildCount;
                
            } else {
                markRange( id, range, child );
            }
        }
    } else {
        r.selectNode(node);
        // r is before range: r.start = r.end < range.start
        if( range.compareBoundaryPoints(Range.END_TO_START,r) > 0 ) 
        {
            return;
        }
        // r is after range: range.end < r.start
        if( range.compareBoundaryPoints(Range.START_TO_END,r) < 0 )
        {
            return;
        }

        $(node).addClass("requality_node id_"+id);

        if( range.compareBoundaryPoints(Range.START_TO_START,r) == 0 )
        {
            $(node).before( $("<a/>").attr("name",id).attr("id","id_"+id).attr("class","requality_id") );
        }
    }
}

/**
 * Mark selection with tags/styles
 * 
 * @param id - location unique id
 * @returns selected xhtml fragment
 */
function markSelection( id ) {
    var selection = window.getSelection();
    switch( selection.rangeCount )
    {
    case 0:
        alert( "Selection is empty" );
        return null;
    case 1:
        break;
    default:
        alert( "Only the first part of selection will be marked. Please, select other parts separately" );
    }
    var range = selection.getRangeAt(0);
    var text = getSerializer().serializeToString( range.cloneContents() );
    var container = range.commonAncestorContainer;
    if( container.nodeType == Node.TEXT_NODE )
    {
    	// range is a part of text node. Mark intersection. 
        markTextRange( id, range, container );
    } else {
    	// Find text node children and mark intersections.
        // Find nodes without text children and mark them also.
        markRange( id, range, container );
    }
    selection.collapseToStart();
    return text;
}

/**
 * This function wrap selection in special requality tags.
 * <br><br><i>Can be used in java</i>
 * @param id UUID
 * @param userId user visible id( <i>For example, 'some.id.01'</i> )
 * @return - selection text
 */
function createLocation( id, reqId ) {
    var text = markSelection( id );
    addLocationLink( id, reqId );
    return text;
}

function addLocationLinks() {
//	var locations = getDocumentLocationIds();
//	if( locations.length == 0 ) return;
//	var id2links = getLocationsToRequirementMap( locations );
//	for( var i = 0; i < id2links.length; i++ )
//	{
//		var chunk = id2links[i];
//		for( var j = 1; j < chunk.length; j++ )
//		{
//			addLocationLink( chunk[0], chunk[j] );
//		}
//	}		
}

function addLocationLink( id, link ) {
    $("#id_"+id).after($("<span>"+link+"</span>").addClass("requality_link"));
}

function removeLocationLink( id, link ) {
    $("#id_"+id).parent().children(".requality_link :contains('"+link+"')").remove();
}

function removeAllLinks() {
    $(".requality_link").remove();
}

function getSelectedLocationUUID() {
    var s = window.getSelection();
    var f = s.focusNode;
    var id = $(f).parents(".requality_text,.requality_node").attr("class");
    if( id == null )
    {
        id = $(f).children(".requality_node").attr("class");
        if( id == null )
        {
//    	    alert( "No selected location" );
            return null;
        }
    }
    var spc = id.indexOf(" ",18);
    if( spc == -1 ) spc = id.length;
    id = id.substring(18,spc);
    return id;
}

function getDocumentLocationIds()
{
	var res = [];
	$(".requality_id").each(
		function()
		{
			var id = this.getAttribute("id");
			res.push( id.substring(3) );
		}
	);
	return res;
}

function removeLocation( id ) {
    $("a.requality_id#id_"+id).remove();
	var t = $("span.requality_text.id_" + id );
	t.children( ".requality_link" ).remove();
	// move text from inside of span before it
	t.each(
		function()
		{
	    	while( this.childNodes.length > 0 )
	    	{
		        this.parentNode.insertBefore( this.childNodes.item(0), this );
		    }
		}
	);
	t.remove();
	$(".requality_node.id_"+id).removeClass("requality_node requality_node_active id_"+id);
}

function deactivateLocations() {
    $( ".requality_text" ).removeClass( "requality_text_active" );
    $( ".requality_node" ).removeClass( "requality_node_active" );
}

function activateLocation( id, scroll ) {
    deactivateLocations();
    
    var location = $( "#id_" + id );
    if( location.size() == 0 )
    {
        alert( "Location not found. id = " + id );
        return;
    }
    
    $( ".requality_text.id_"+id ).addClass( "requality_text_active" );
    $( ".requality_node.id_"+id ).addClass( "requality_node_active" );
    
    if( scroll )
    {
        window.scrollTo( 0, location.offset().top - window.innerHeight / 2 );
    }
}

function activateSelectedLocation() {
    var id = getSelectedLocationUUID();
    if( id )
    {
        activateLocation( id, false );
    } else {
        deactivateLocations();
    }
}

/**
 * Search for aString with given options.
 * <br><br><i>Can be used in java</i>
 */
function findText( aString
                 , aCaseSensitive
                 , aBackwards
                 , aWrapAround
                 )
{
    return window.find( aString
                      , aCaseSensitive
                      , aBackwards
                      , aWrapAround
                      , false
                      , false
                      , false
                      );
}

/**
 * Selecting requality Location node and makes it orange.
 * <br><br><i>Can be used in java</i>
 */
function selectNode() {
    alert( 'selectNode' );
}

function getBrowserInfo() {
	var appName= navigator.appName;
    var appVersion= navigator.userAgent;
    return "Browser: " + appName + "; Version: " + appVersion;
}

/**
 * <i>Can be used in java</i>
 * @return - <b>true</b> if selection can be wrapped by DOM element<br>
 * - and <b>false</b> otherwise.   
 */
function isSelectable() {
//    alert( 'isSelectable' );
    var s = window.getSelection().toString();
    return s.length>0;
}

/**
 * <i>Can be used in java</i>
 * @return - <b>true</b> if selected node can be unwrapped
 */
function isRemovable() {
    return getSelectedLocationUUID() != null;
}

document.onclick =
    function( e )
    {
        // right mouse button
        if( e.button == 2 )
        {
            var s = window.getSelection();
            var n = s.rangeCount;
            for( var i = 0; i < n; i++ )
            {
                var r = s.getRangeAt(i);
                // if one of selection ranges contains current mouse position - do nothing
                // WARNING: isPointInRange is mozilla specific function
                if( r.isPointInRange( e.rangeParent, e.rangeOffset ) ) return;
            }
            // collapse selection to current mouse position
            s.collapse( e.rangeParent, e.rangeOffset );
            activateSelectedLocation();
        }
    };

function propNames( obj )
{
    if( !obj ) return "obj is undefined";
    var s = "/*" + obj.toString() + ":*/{";
    var comma = false;
    for( p in obj ) { 
        if( comma ) s += ",";
        comma = true;
        s += "|" + p + "=" + obj[p];
    }
    s += "\n}";
    return s;
}

function inspect( obj ) {
    alert( propNames( obj ) );
}

function inspectRange( tip, range )
{
    var text = getSerializer().serializeToString( range.cloneContents() );
    alert( tip + text );
}

function inspectNode( tip, node )
{
    var range = document.createRange();
    range.selectNode(node);
    inspectRange( tip, range );
}

function inspectRangeNode( tip, range, node )
{
    var rangeText = getSerializer().serializeToString( range.cloneContents() );

    var nodeRange  = document.createRange();
    nodeRange.selectNode(node);
    var nodeText = getSerializer().serializeToString( nodeRange.cloneContents() );

    alert( tip + " <" + rangeText + "> <" + nodeText + ">" );
}

function tree( node )
{
	var s = "";
	switch( node.nodeType )
	{
	case Node.ELEMENT_NODE:
		s += "<"+node.nodeName;
		s += " class=" + node.getAttribute("class") + ">";
		for( var i = 0; i < node.childNodes.length; i++ )
		{
			s += "\n"+tree( node.childNodes.item(i) );
		}
		s += "\n</"+node.nodeName+">";
		break;
	case Node.TEXT_NODE:
		s += node.nodeValue;
	}
	return s;
}	

addLocationLinks();
setInitialLocation();
